/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <math.h>

/* Obliczanie cisnienia w funkcji wysokosci */
/* Ponizsze stale definiuja cechy planety i jej atmosfery; jesli chcesz
   przewidywac widzialnosc na Marsie, musisz je zmienic ;) */
/* Srednia masa molowa powietrza, w kg/mol */
#define M_MOL	0.0289644
/* Stala gazowa w J/(mol K); ta przynajmniej wszedzie jest taka sama */
#define R_G	8.31447
/* Stala czesc wspolczynnika proporcjonalnosci we wzorze barometrycznym */
#define C_P	M_MOL / R_G
/* Stale w funkcji przyspieszenia od szerokosci geograficznej i wysokosci:
   g=G_C*(1+CG_1*sin^2(t)-CG_2*sin^2(2t))-CG_H*h */
#define G_C	9.780318
#define CG_1	5.3024e-3
#define CG_2	5.8e-6
#define CG_H	3.086e-6
/* Wymagana dokladnosc obliczen cisnienia */
#define P_EPS	0.01

/* Zmienne lokalne dla tego pliku: ostatnio rozwazana wysokosc
   i obliczone (lub zdefiniowane z zewnatrz) na niej cisnienie. */

static double hl = 0.;
static double pl = 101325.;	/* Wartosci domyslne */

extern void temph();
extern void blad();
extern int debug, verb;
extern double b;	/* Polos biegunowa */

/* Odczytanie i ustawienie bieacego cisnienia i wysokosci */
void pget(double *h, double *p) {
	*h = hl;
	*p = pl;
}

void pset(double h, double p) {
	hl = h;
	pl = p;
}

/* Wczytanie wartosci cisnienia na poziomie morza */
void pinit() {
	if (verb)
		printf("Podaj wysokosc (w m) i cisnienie na niej (w hPa)\n");
	scanf("%lg %lg",&hl,&pl);
	if (pl < 0.)
		blad("Cisnienie nie moze byc ujemne.");
	pl *= 100.;
}

/* Obliczenie wartosci cisnienia, pochodnej cisnienia
   temperatury i pochodnej temperatury po wysokosci. */
/* Wzor barometryczny: dln(p)/dh = - C_P * g / T
   gdyby iloraz g / T nie zalezal od h, moglibysmy to scalkowac analitycznie
   do p(h2)=p(h1) exp(-C_P * g * h / T) - ale na ogol tak nie jest. */
/* h jest wskaznikiem do tablicy, w ktorej jest takze odleglosc od srodka
   elipsoidy, szerokosc i dlugosc geograficzna. */
void ptwart(double *h, double *pt) {
	double hp, dh, pp, ptmp, dht, tdh;
	double g, gc;	/* Do obliczen lokalnego przyspieszenia */
	double cp;
	double sg, sg2;
	double *th;
	int i, n;
/* Sinusy szerokosci geograficznej i podwojonej szerokosci geograficznej,
   do wzoru na lokalne przyspieszenie grawitacyjne */
	sg = sin(h[2]);
	sg2 = sin(h[2]+h[2]);
	gc = G_C * (1. + CG_1*sg*sg - CG_2*sg2*sg2);
	/* Obliczamy nowe cisnienie, przyjmujac stala wartosc pochodnej
	   temperatury po wysokosci w rozwazanym przedziale. Potem
	   tniemy ten przedzial na mniejsze kawalki i sprawdzamy dokladnosc. */
	/* Tablica temperatur to po prostu 3-ci i 4-ty element tablicy pt. */
	th = pt + 2;
	pt[0] = pl;
	if (*h == hl)
		goto pochodna;
	/* Zapamietujemy docelowa wysokosc, zeby moc wywolac temph
	   ze wspolrzednymi geograficznymi (moze na przyszlosc). */
	hp = *h;
	/* Zaczynamy od jednego punktu */
	n = 1;
powtorz:
	pp = pl;
	h[0] = hl;
	dh = (hp - hl);
	/* h[1] -= dh; Korygujemy odleglosc od srodka elipsoidy - niepoprawne, jesli mamy prawdziwy pion */
	dh /= n;
	for (i = 0; i < n; i++) {
		/* Obliczamy lokalna wartosc przyspieszenia grawitacyjnego */
		g = gc - CG_H * h[0];
		/* Wspolczynnik do obliczen: masa molowa powietrza * przyspieszenie / stala gazowa */
		cp = C_P * g;
		/* potrzebujemy temperatury i jej pochodnej */
		temph(h, th);
		dht = dh / th[0];
		tdh = th[1] * dht;
		if (fabs(tdh) < 1.e-16)
			pp *= exp(-cp * dht);
		else
			pp *= pow(1.+tdh,-cp/th[1]);
		if (debug > 1)
			printf("h=%lf t=%lf dt/dh=%lf p=%lf, g=%.6lf\n",h[0],th[0],th[1],pp,g);
		h[0] += dh;
		/* h[1] += dh; Korekta odleglosci od srodka elipsoidy - patrz wyzej */
	}
	/* Spradzenie zbieznosci */
	ptmp = pt[0];
	pt[0] = pp;
	if (fabs(pp - ptmp) > P_EPS) {
		n += n;
		goto powtorz;
	}
	hl = h[0];
	pl = pt[0];
pochodna:
	temph(h, th);
	g = gc - CG_H * h[0];
	cp = C_P * g;
	tdh = cp / th[0];
	pt[1] = -tdh * pl;
	if (debug)
		printf("p i T na wysokosci h=%lf p=%lf, dp/dh=%lf, T=%lf, dT/dh=%lf, g=%lf\n",hl,pt[0],pt[1],th[0],th[1],gc-CG_H*hl);
}
