/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <stdlib.h>

/* Interpolacja funkcji tzw. funkcjami sklejanymi 3-go stopnia.
   Funkcja w przedziale <x[k]-x[k+1]> ma postac:
   f(r) = a[k]*h^3 + b[k]*h^2 + c[k]*h + f[k], gdzie h = x[k+1] - x[k],
   dla x > x[np-1] lub x < x[0] zakladamy zaleznosc liniowa.
 */

static int np = 0; /* liczba punktow */
static double *x,*f,*a,*b,*c; /* punkty i wspolczynniki */
/* UWAGA! x i f wskazuja na pamiec zaalokowana na zewnatrz
   tego modulu programu! NIE WOLNO jej zwalniac przed zakonczeniem
   wszelkich interpolacji wykorzystujacych fukcje sklejane poprzez
   funkcje fpint! To nie jest eleganckie rozwizanie i jeli zajdzie
   potrzeba, to ten kod powinien by przerobiony tak, by przechowywa
   lokalne kopie. */

extern void blad(char *);
extern char * xmalloc();


/* Rozwiazanie ukladu rownan o strukturze trojprzekatniowej:
   a - tablica wspolczynnikow (tylko niezerowe przy tej strukturze!)
   b - tablica wyrazow wolnych; na wyjsciu rozwiazania */
static int urtp(double *a, double *b, int n) {
	double x;
	double *a1, *a2;
	int w;
	/* Zerowanie wspolczynnikow pod przekatna */
	a1 = a;
	a2 = a1 + 2;
	for (w = 0; w < n-1; w++) {
		x = (*a2)/(*a1);
		/* elementu pod przekatna nie bedziemy i tak potrzebowac
		   - po co tracic czas na jego jawne zerowanie? */
		a1++;
		a2++;
		/* Modyfikacja elementu lezacego na przekatnej w rzedzie w+1 */
		*a2 -= x*(*a1);
		/* Wyraz wolny */
		b[w+1] -= x*b[w];
		/* Modyfikacja wskaznikow powinna byc (chyba) szybsza
		   od operacji zwiazanych z indeksowaniem tablicy a */
		a1 = a2;
		a2 = a2 + 2;
	}
	/* W tym miejscu pod przekatna mamy zera - mozna rozwiazywac */
	/* a1 wskazuje na ostatni element na przekatnej */
	b[n-1] /= (*a1);
	for (w = n-2; w >= 0; w--) {
		a2 = a1 - 2;
		a1 = a2 - 1;
		b[w] = (b[w] - b[w+1]*(*a2))/(*a1);
	}
	return 0;
}

/* Obliczenie wartosci fnkcji i jej pochodnej dla danego x */
void fpint(double xx, double *fp) {
	double h;
	static int i = 0;
	/* Zakladamy, ze wywolania sa lokalne - zapamietujemy aktualny
	   przedzial wysokosci i w kolejnym wywolaniu zaczynamy od niego. */
	if (xx < x[i]) {	/* Idziemy w strone mniejszych wysokosci */
		for (i--; i >= 0; i--)
			if (xx >= x[i])
				goto przedzial;
		i = 0;
		goto skrajny;
	} else {		/* Idziemy w strone wiekszych wysokosci */
		for (i++; i < np; i++)
			if (xx <= x[i]) {
				i--;
				goto przedzial;
			}
		i = np - 1;
		goto skrajny;
	}
przedzial:	/* Punkt jest miedzy x[i] a x[i+1]. */
	h = xx - x[i];
	/* printf ("%d: a=%lg b=%lg c=%lg d=%lg h=%lg\n",i,a[i],b[i],c[i],f[i],h); */
	fp[0] = ((a[i]*h + b[i])*h +c[i])*h + f[i];
	fp[1] = (3*a[i]*h + 2*b[i])*h + c[i];
	return;
skrajny:
	h = xx - x[i];
	fp[0] = f[i] + c[i] * h;
	fp[1] = c[i];
}

/* Obliczenie wartosci wspolczynnikow wielomianow 3-go stopnia
   w poszczegolnych przedzialach; podane sa wartosci argumentu i funkcji
   w n punktach oraz pierwsza pochodna funkcji w punktach skrajnych. */
void splinit(double *xx, double *fx, int n, double fp1, double fpn) {
	int i;
	double h0, h1, hh, d0, d1;
	double *aa, *atmp;
	if (np != 0)	/* To nie musi byc pierwsze i jedyne wywolanie! */
		free(a);
	np = n;
	x = xx;		/* Zapisujemy wskazniki do argumentow */
	f = fx;		/* i wartosci funkcji; beda potrzebne do interpolacji! */
	if (np < 3)
		blad("Interpolowana funkcja musi byc podana przynajmniej w 3 punktach.");
	a = (double *) malloc(sizeof(double)*(3*np-2));
	b = a + np - 1;
	c = b + np - 1;	/* W c[np-1] zapamietamy pochodna t po r w ostatnim punkcie */
/* Sprawdzenie kolejnosci punktow; argumenty musza tworzyc ciag rosnacy! */
	for (i = 1; i < np; i++)
		if (x[i] <= x[i-1])
			blad("Punkty do interpolacji funkcji nie sa posortowane.");
/* Wspolczynniki d sa po prostu rowne wartosciom funkcji */
	/* Zaalokowanie pamieci na rozwiazanie ukladu rownan */
	aa = (double *) xmalloc(sizeof(double)*(3*np-8));
	/* Ukladamy rownania na wspolczynniki c */
	c[0] = fp1;	/* Wartosc pochodnej w pierwszym punkcie */
	i = 1;
	h0 = x[1] - x[0];
	h1 = x[2] - x[1];
	d0 = f[1] - f[0];
	d1 = f[2] - f[1];
	atmp = aa;
	hh = h1/h0;
	goto wskocz;
nast:
	i++;
	h0 = h1;
	d0 = d1;
	h1 = x[i+1] - x[i];
	d1 = f[i+1] - f[i];
	hh = h1/h0;
	atmp++;
	*atmp = hh;
	atmp++;
wskocz:
	*atmp = 2 * (hh+1.);
	c[i] = 3*(hh*d0/h0+d1/h1);
	if (i == np-2) goto koniec;
	atmp++;
	*atmp = 1.;
	goto nast;
koniec:
	/* Uwzgledniamy pochodne na poczatku i na koncu przedzialu */
	c[1] -= (x[2]-x[1])*fp1/(x[1]-x[0]);
	c[np-2] -= fpn;
	c[np-1] = fpn;
	/* Rozwiazanie ukladu rownan */
	urtp(aa, &(c[1]), np-2);
	/* W tym miejscu mamy juz wspolczynniki c wielomianu */
	/* Liczymy wspolczynniki b */
	for (i = 1; i < np-1; i++) {
		h0 = x[i]-x[i-1];
		b[i] = (c[i-1] + 2*c[i] - 3*(f[i]-f[i-1])/h0)/h0;
	}
	b[0]=-b[1]+(c[1]-c[0])/(x[1]-x[0]);
	/* I wreszcie wspolczynniki a */
	for (i = 0; i < np-1; i++) {
		h0 = x[i+1] - x[i];
		a[i] = (((f[i+1] - f[i])/h0 - c[i])/h0 - b[i])/h0;
		/* printf ("i=%d a=%lg b=%lg c=%lg\n",i,a[i],b[i],c[i]); */
	}
	free(aa);
}
