/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <math.h>
#include "mw3.h"

extern void piszp();
extern void g2k();
extern void k2g();
extern void ptwart();
extern double relips();
extern void macierz_obrotu();
extern double pion();
extern void pset();
extern void pget();

extern int debug, verb;

/* Zmienne, w ktorych przechowywana jest aktualna najmniejsza wysokosc,
   na ktorej rozchodzi sie swiatlo i odleglosc tego miejsca od punktu
   poczatkowego. */
static double rmin, hmin;

/* Stale do obliczania wspolczynnika zalamania swiatla. Odnosza sie do
   atmosfery ziemskiej - jesli chcesz przewidywac widocznosc na Marsie,
   gdzie gory sa wyzsze, musisz je najpierw zmienic ;)
   W obliczeniach wykorzystywany jest model Lorentza polaryzowalnosci
   dielektryka. Wspolczynnik zalamania swiatla jest rowny pierwiastkowi
   ze wzglednej przenikalnosci elektrycznej osrodka. Polaczenie modelu
   Lorentza z rownaniem stanu gazu doskonalego prowadzi do zaleznosci
   n=sqrt((1+2*C_N*(p/T))/(1-C_N*(p/T))), gdzie C_N=a/(k e_0) i jest obliczone
   ponizej. Symbol a oznacza srednia polaryzowalnosc skladnikow powietrza,
   k - stala Boltzmanna, e_0 - przenikalnosc elektryczna prozni,
   p - cisnienie (w Pa), T - temperature (w K). Poniewaz dokladne wartosci
   polaryzowalnosci tlenu i azotu dla pol elektrycznych o czestotliwosci
   odpowiadajacej falom swietlnym nie sa znane z wystarczajaca dokladnoscia,
   wyrazenie C_N=a/(k e_0) jest traktowane jako parametr modelu i dopasowane
   do znanej wartosci n=1.0002926 pod cisnieniem standardowym 101325 Pa
   i w temperaturze 273.15 K.
*/
#define N_REF	1.0002926
#define T_REF	273.15
#define P_REF	101325.
#define NN	N_REF*N_REF
#define C_N	((NN-1.)*T_REF)/((NN+2.)*P_REF)

/* Obliczenie pochodnej logarytmu naturalnego wspolczynnika zalamania swiatla
   po lokalnym kierunku pionowym. 
   Wspolczynnik zalamania swiatla n zalezy
   od cisnienia i temperatury. Wartosci p i T oraz ich pochodne
   po kierunku pionowym obliczane sa w funkcji ptwart(). Przyjmujemy,
   ze pochodne p i T w kierunkach poziomych sa zaniedbywalnie male.
   Korzystamy ponadto z zaleznosci d(ln n)/dx = (1/n)(dn/dx) */

/* tablica hrsd zawiera wysokosc nad poziomem morza
   i wspolrzedne geograficzne punktu;
   v jest wektorem pionu; w przyszlosci funkcja niniejsza moze go modyfikowac,
   tak by wzdluz niego najszybciej malal wspolczynnik zalamania swiatla.
   Wtedy nie bedzie juz pionem i dopuszczalne bedzie poziome odchylenie
   promienia swiatla. */

static double derlnn(double *hrsd, double *v) {
	double pt[4];	/* cisnienie, temperatura i ich pochodne) */
	double wn, wp;	/* wspolczynnik zalamania swiatla i jego pochodna */
	double dpt, w1, w2;	/* zmienne na iloraz p/T i inne wyrazenia robocze */

/* Obliczenie cisnienia, temperatury i ich pochodnych po wysokosci */
	ptwart(hrsd,pt);
	dpt = pt[0] / pt[2];
	w2 = C_N * dpt;
	w1 = 1. - w2;
	w2 += 1. + w2;
	w2 /= w1;
	wn = sqrt(w2);
	wp = (0.5 * wn + 1. / wn) * (C_N / (pt[2] * w1)) * (pt[1] - dpt * pt[3]);
	w1 = wp / wn;
	if (debug) {
		printf("derlnn: p=%lg dp/dh=%lg T=%lg dT/dh=%lg\n",pt[0],pt[1],pt[2],pt[3]);
		printf("h=%.3lf, n=%.8lf, dn/dh=%.6lg, dln(n)/dh=%.6lg\n",hrsd[0],wn,wp,w1);
	}
	return w1;
}

/* Funkcje wyznaczajace trajektorie promienia swiatla wedlug roznych modeli */

/* Obliczenie wektora przesuniec stosowanego w symulacjach;
   zwraca tez odleglosc miedzy punktami. */
static double optdx(double *x1, double *x2, double *dx, double *s) {
	double xd;
/* Obliczenie odleglosci pomiedzy punktami */
	aminusb3(x2, x1, dx);
	xd = sqrt(adotb3(dx, dx));
/* Czy podany krok nie jest zbyt dlugi? */
	if  (*s >= xd - 1.)
		*s = xd;
/* Skalowanie s - bedzie dalej sluzyc jako kryterium dokladnosci */
/*	*s /= xd; */
/* Tworzymy wektor przesuniecia */
	dscal3(dx, (*s)/xd);
	return xd;
}

/* Wydrukowanie opisu punktu; dane sa wspolrzedne kartezjanskie
   punktow biezacego i poczatkowego. */
static void piszpkart(double *x, double *xp, int setmin) {
	double rsdh[4], dx[3], r;
	double eps = 0.001;
	/* Obliczenie wspolrzednych geograficznych i wysokosci
	   nad poziomem morza biezacego punktu. */
	rsdh[3] = pion(x, rsdh, dx, eps);
	/* Obliczenie odleglosci od punktu poczatkowego */
	aminusb3(xp, x, dx);
	r = sqrt(adotb3(dx, dx));
	/* Zapamietanie najmniejszej odleglosci */
	if (setmin || rsdh[3] < hmin) {
		rmin = r;
		hmin = rsdh[3];
	}
	/* Wydrukowanie wspolrzednych i wysokosci nad poziomem morza punktu */
	if (verb)
		piszp(r, rsdh + 1);
}

/* Rozchodzenie swiatla po liniach prostych. Wektor w ma 6 elementow,
   zwracane w nim sa odleglosc miedzy punktami, wysokosc minimalna,
   odleglosc od x1, w ktorej osiagana jest wysokosc minimalna i wektor
   opisujacy kierunek, w ktorym widziany jest punkt docelowy x2. */
void prop_prosto(double *x1, double *x2, double s, double *w) {
	double x[3], xr[3], r2;
	double *dx;
	int setmin = 1;
	dx = w + 3;
	w[0] = optdx(x1, x2, dx, &s);
	
/* Skopiowanie wspolrzednych punktu poczatkowego */
	dcopy3(x1, x);
	do {
		piszpkart(x, x1, setmin);
		setmin = 0;
/* Polozenie nastepnego punktu */
		aplusb3(x, dx, x);
/* Obliczenie odleglosci punktu biezacego od punktu docelowego. */
		aminusb3(x, x2, xr);
		r2 = sqrt(adotb3(xr, xr));
	} while (r2 > s + 1.);
	piszpkart(x, x1, setmin);
	if (r2 > 1.) {
		aplusxb3(x, dx, x, r2/s);
		piszpkart(x, x1, setmin);
	}
	w[1] = hmin;
	w[2] = rmin;
}

/* Obliczenia trajektorii promienia swiatla (pojecie z optyki geometrycznej,
   chociaz rachunki opieraja sie na zasadzie Huyghensa, nalezacej do optyki
   falowej). Do obliczen potrzebna jest wartosc pochodnej logarytmicznej
   wspolczynnika zalamania swiatla po kierunku prostopadlym do biezacego
   kierunku rozchodzenia promienia swiatla. Obliczane jest polozenie kolejnego
   punktu na trajektorii i nowy wektor kierunku rozchodzenia sie swiatla.

   Rownowazny wynik powinno sie otrzymac, nie wychodzac poza optyke
   geometryczna (z zasady Fermata).

   Wszystkie dane i wyniki sa w globalnym, kartezjanskim ukladzie wspolrzednych,
   ktorego poczatek pokrywa sie ze srodkiem elipsoidy, zas os z przechodzi
   przez bieguny.
   x - poczatkowy punkt na trajektorii promienia swiatla
   w - wektor biezacego kierunku rozchodzenia sie swiatla; jego dlugosc
       okresla, jak dalego ma byc docelowy punkt na wykonywanym etapie obliczen
   xn - nowy punkt na trajektorii
   wn - nowy wektor kierunku rozchodzenia sie swiatla
*/
static void nastpunkt(double *x, double *w, double *xn, double *wn, double eps) {
	double os[3], rm[9], xw[12], p[3], hrsd[4];
	double dlnn, wl, xl, alfa, beta, dnw, dnw2, xs, epsn, pl;
	double *x1, *x2, *w1, *w2, *tmp;
	int i, n;

/* Wskazniki do roboczych obszarow pamieci */
	x1 = &xw[0]; x2 = x1 + 3; w1 = x2 + 3; w2 = w1 + 3;
/* Kopiujemy punkt poczatkowy do koncowego, zeby test zbieznosci na pewno
   nie powiodl sie po pierwszym przebiegu */
	dcopy3(x, xn);
	dcopy3(w, wn);
	n = 1;	/* Liczba podprzedzialow */
powtorz:
	if (debug)
		printf("Podzial modelowanego odcinka na %d czesci.\n",n);
/* Skopiowanie punktu startowego i wektora kierunku do tablic lokalnych */
	dcopy3(x, x1);
	dcopy3(w, w1);
/* Skalowanie */
	dscal3(w1, 1./(double) n);
	epsn = eps/n;
/* Petla po podprzedzialach */
	for (i = 0; i < n; i++) {
/* Wyznaczenie lokalnego pionu (p) */
		hrsd[0] = pion(x1, hrsd+1, p, epsn);
/* Obliczenie pochodnej logarytmicznej wspolczynnika zalamania swiatla
   po kierunku lokalnego pionu w punkcie x1 */
		dlnn = derlnn(hrsd, p);
/* Wektor p wyznaczal lokalny pion, ale w derlnn mogl zostac zmodyfikowany
   i teraz wyznacza wektor, wzdluz ktorego najszybciej maleje wspolczynnik
   zalamania swiatla; obliczamy kat pomiedzy w1 a p */
		wl = sqrt(adotb3(w1, w1));
		pl = sqrt(adotb3(p, p));
		alfa = acos(adotb3(w1, p) / (wl * pl));
/* Rzutujemy dlnn na kierunek prostopadly do kierunku rozchodzenia sie
   swiatla (zakladamy, ze pochodna po lokalnym kierunku poziomym jest
   rowna zeru). Wykorzystana jest rownosc sin(alfa) = cos(alfa - pi/2) */
		dlnn *= sin(alfa);
/* Obliczamy kat beta, o jaki odchyli sie promien swiatla po pokonaniu
   odleglosci wl i odchylenie y w kierunku prostopadlym do wyjsciowego
   kierunku rozchodzenia sie swiatla. Zastosowane jest przyblizenie roznic
   skonczonych. */
	   	dnw = dlnn * wl;
	   	beta = atan(dnw);
	   	if (debug)
   			printf("dlnn/dp=%lg, alfa=%lg, beta=%lg\n",dlnn,alfa*180./M_PI,beta*180./M_PI);
/* Decydujemy, czy uzyc wzoru przyblizonego - dobrego dla malych katow beta
   - czy dokladnego, lecz niestabilnego numerycznie dla malych katow. */
		dnw2 = dnw * dnw;
		if (dnw2 < 1.e-8) {
			xs = 0.5 * dnw * wl;
			if (debug)
				printf("wzor przyblizony, przesuniecie=%lg\n",xs);
		} else {
			xs = (sqrt(1. + dnw2) - 1.) / dlnn;
			if (debug)
				printf("wzor dokladny, przesuniecie=%lg\n",xs);
		}
/* Obliczamy os obrotu do pozniejszej korekty kierunku rozchodzenia sie
   swiatla, jako iloczyn wektorowy w1 i p - wektor prostopadly do w1 i p */
		axb3(w1, p, os);
/* Znajdujemy wektor prostopadly do w1, lezacy w plaszczyznie w1-p;
   korzystamy z wektora osi obrotu, obliczonego przed chwila */
		axb3(os, w1, p);
/* Obliczamy nowy wektor polozenia - najpierw przesuniecie w biezacym kierunku */
		aplusb3(x1, w1, x2);
/* a potem w prostopadlym do niego; */
		aplusxb3(x2, p, x2, xs / sqrt(adotb3(p, p)));
/* Macierz obrotu wokol znalezionej wczesniej osi o kat beta */
		macierz_obrotu(rm, os, beta);
/* Obracamy wektor, co daje nowy wektor kierunku rozchodzenia */
		mawek3(rm, w1, w2);
/* Zamieniamy wskazniki */
		tmp = x1; x1 = x2; x2 = tmp; tmp = w1; w1 = w2; w2 = tmp;
	}
/* Test zbieznosci - liczymy odleglosc pomiedzy punktami koncowymi i dlugosc
   roznicy wektorow kierunku, otrzymanych przy roznych dlugosciach kroku
   symulacji. */
	aminusb3(x1, xn, x2);
	xl = sqrt(adotb3(x2, x2));
	dscal3(w1, (double) n);
	aminusb3(w1, wn, w2);
	wl = sqrt(adotb3(w2, w2));
	dcopy3(x1, xn);		/* To na wypadek, */
	dcopy3(w1, wn);		/* gdyby zbieznosc zostala osiagnieta. */
	if (debug)
		printf("n=%d, eps=%.6lf, xl=%.6lf, wl=%.6lf\n",n,eps,xl,wl);
	if (xl > eps || wl > eps) {
		n += n;
		goto powtorz;
	}
}

/* Ta funkcja uwzglednia uginanie swiatla w osrodku o zmiennym 
   wspolczynniku zalamania; dane - wspolrzedne kartezjanskie punktow
   startowego i docelowego oraz odleglosc, jaka ma byc pokonana
   w pojedynczym kroku. Dokladnosc symulacji wynosi 1 m.
   Wektor w zawiera wyniki, tak jak w prop_prosto(). */
void prop_ugiecie(double *x1, double *x2, double s, double *w, int skan, int *jest, double hdoc) {
	double dx[3], xdx[12], rot[3], rm[9];
	double r12, r2, x2x2, p12, eps, tgak;
	double tga[2], tgai, alfa, dh[2], dhi;
	double htmp, ptmp;
	double *xp1, *xp2, *dp1, *dp2, *t;
	static double tgap, dhp, difh, sumh;
	int n, setmin, bskan, bylo;
/* Liczba punktow, ktore juz mamy */
	n = 0;
	bskan = 0;
	x2x2 = adotb3(x2, x2);	/* bedzie potrzebne pozniej */
	r12 = optdx(x1, x2, dx, &s);
	w[0] = r12;
	eps = 0.5 * s / r12;	/* dokladnosc w wewnetrznym kroku obliczen */
	if (debug)
		printf("Wymagana dokladnosc obliczen w symulacji wynosi %lg\n",eps);
/* Znajdujemy os, wokol ktorej bedziemy obracac wewnetrzny wektor */
	axb3(dx, x1, rot);
/* Wskazniki do tablic roboczych; jest tak smiesznie, bo bedziemy je przestawiac */
	xp1 = &xdx[0]; xp2 = xp1 + 3; dp1 = xp2 + 3; dp2 = dp1 + 3;
/* skopiowanie wektora przesuniecia */
	dcopy3(dx, dp1);
	if (skan) {
		bylo = 0;
		bskan = 1;
		if (*jest) {
			if (verb)
				printf("Kontynuacja skanowania, tgap = %lf\n",tgap);
			tgai = tgap + tgak;
			goto obrot;
		}
		tgak = (double) skan / r12;
		tgai = 0.;
		tgap = 0.;	/* Inicjalizacja tangensa kata obrotu */
		dhp = 0.;
		if (verb)
			printf("Tangens kata patrzenia bedzie zmieniany z krokiem %lf\n",tgak);
	}
model:
	if (verb)
		printf("Poczatek symulacji\n");
	/* Zachowanie biezacej wysokosci i cisnienia, zeby nastepny
	   przebieg symulacji wystartowal z tych samych wartosci. */
	pget(&htmp, &ptmp);
	/* Kopia do tablicy wynikow */
	dcopy3(dp1, w+3);
	setmin = 1;
/* Skopiowanie wspolrzednych punktu poczatkowego */
	dcopy3(x1, xp1);
	do {
		piszpkart(xp1, x1, setmin);
		setmin = 0;
/* Obliczenie nastepnego punktu */
		nastpunkt(xp1, dp1, xp2, dp2, eps);
		/* Zamiana punktow i wektorow */
		t = xp1; xp1 = xp2; xp2 = t;
		t = dp1; dp1 = dp2; dp2 = t;
/* Obliczenie odleglosci punktu biezacego od wektora pionu punktu docelowego. */
/* Obliczamy wektor od punktu biezacego do wektora x2 */
		aplusxb3(xp1, x2, dp2, -adotb3(xp1, x2)/x2x2);
		r2 = sqrt(adotb3(dp2, dp2));
	} while (r2 > s + 1. && r2 <= r12);	/* ten drugi test jest "na wszelki wypadek" */
	piszpkart(xp1, x1, setmin);
/* Ostatni etap wymaga dopasowania wektora przesuniecia */
/* Obliczamy wektor od punktu biezacego do wektora x2: */
	dcopy3(x2, xp2);
/* najpierw rzut biezacego punktu na wektor x2 */
	dscal3(xp2, adotb3(xp1, xp2) / x2x2);
	aminusb3(xp2, xp1, dp2);
/* W dp2 mamy wektor prostopadly do x2, od xp1 do tego wektora. */
	r2 = adotb3(dp2, dp2);
	p12 = adotb3(dp1, dp2);
	if (r2 > 1. && fabs(p12) > 0.) {
/* Chcemy, zeby wektor dp1 siegal do wektora x2. */
		dscal3(dp1, r2 / p12);
		nastpunkt(xp1, dp1, xp2, dp2, eps);
/* Obliczenie odleglosci punktu biezacego od punktu poczatkowego. */
		piszpkart(xp2, x1,setmin);
	}
/* Sprawdzamy roznice wysokosci punktu docelowego i koncowego punktu symulacji */
 	k2g(x2, dp1);
	k2g(xp2, dp2);
	dhi = dp1[0] - dp2[0];
	if (verb)
		printf("Roznica wysokosci pomiedzy punktem docelowym a osiagnietym dh = %.4lf\n",dhi);
	if (skan) {
		if (bskan) {
			if (bylo && dhp * dhi < 0.) {	/* Znaleziona zmiana znaku! */
				dh[0] = dhp;
				tga[0] = tgap;
				dh[1] = dhi;
				tga[1] = tgai;
				n = 2;
				bskan = 0;
				if (verb)
					printf("Znaleziono przedzial: dh0=%lf, dh1=%lf, tga0=%lf, tga1=%lf\n",dhp,dhi,tgap,tgai);
				goto interp;
			}
			n = 0;
			tgap = tgai;
			dhp = dhi;
			tgai = tgai + tgak;
			if (verb)
				printf("krok skanowania, tga=%lf\n",tgai);
			bylo = 1;
		}
		if (dhi < -hdoc) {	/* W trybie skanowania musimy gdzies skonczyc. */
			*jest = 0;
			if (verb)
				printf("Koniec skanowania\n");
			return;
		}
	} else
		tgak = tgai;
	/* Przywrocenie wyjsciowej wysokosci i temperatury */
	pset(htmp, ptmp);
obracaj:
	if (fabs(dhi) >= 1.e-1) {
/* Oszacowanie kata obrotu poczatkowego wektora kierunku rozchodzenia sie swiatla */
		switch (n) {
		case 0:		/* Jestesmy po pierwszym podejsciu */
			if (!bskan)
				tgai = dhi / r12;
			tga[0] = 0.;
			dh[0] = dhi;
			goto obrot;
		case 1:		/* Szukamy przedzialu kata rotacji */
			if (dhi * dh[0] < 0) {	/* Przeszorowalismy w druga strone */
				dh[1] = dhi;
				tga[1] = tgai;
				goto interp;
			} else {	/* Musimy dalej zmieniac kat w tym samym kierunku */
				dh[0] = dhi;
				tga[0] = tgai;
				tgai += tgak;
				n--;	/* W nastepnym kroku chcemy znowy wyladowac tu */
				goto obrot;
			}
		default:	/* Dalsze proby */
			if (dhi * dh[0] < 0.) {	/* Szukany, optymalny kat jest miedzy dh[0] a dhi */
				dh[1] = dhi;
				tga[1] = tgai;
			} else  {
				dh[0] = dhi;
				tga[0] = tgai;
			}
		}
/* Interpolacja kata obrotu */
interp:		/* Interpolacja funkcja liniowa dh = a*tga + b */
		difh = fabs(tga[1] - tga[0]);
		sumh = fabs(tga[1]) + fabs(tga[0]);
		if (difh <= 1.e-6 * sumh) {
			printf("# Brak stabilnych rozwiazan: dh0 = %.8lg, dh1 = %.8lg, tga0 = %.8lg, tga1 = %.8lg\n",dh[0],dh[1],tga[0],tga[1]);
			goto wyjscie;
		}
		tgai = (dh[1]*tga[0] - dh[0]*tga[1]) / (dh[1] - dh[0]);
/* Obrot wektora poczatkowego */
obrot:
		alfa = atan(tgai);
		if (verb)
			printf("Rotacja wektora kierunku o %.6lg stopni.\n",alfa*180./M_PI);
		macierz_obrotu(rm, rot, alfa);
		mawek3(rm, dx, dp1);
		n++;
		goto model;
	}
	/* Zapisanie wysokosci minimalnej to tablicy wynikow. */
wyjscie:
	w[1] = hmin;
	w[2] = rmin;
	*jest = 1;
}
