/* Copyright (C) Krzysztof Strasburger, 2009
   licencja GNU GPL, wersja 2 lub pniejsza */

#include <stdio.h>
#include <math.h>
#include "mw3.h"

extern void g2k();
extern void splinit();
extern void fpint();
extern char *xmalloc();
extern int debug, verb;

/* Parametry funkcji przesuniecia wysokosci */
static double xref[3], d0, a0;
static double hs = 0.;
static int ndh;

/* Modele zaleznosci temperatury od wysokosci */
static int tmodel = 1;
static double t0 = 273.15;
static double tgrad = -0.0065;

/* 1 - podajemy temperature na wysokosci 0 i pionowy gradient;
   2 - podajemy gradient w dolnej czesci atmosfery, liczbe punktow,
       w ktorych beda podane wartosci temperatury, gradient w gornej
       czesci atmosfery i pary wartosci wysokosc - temperatura.
       W wewnetrznym obszarze temperatura jest opisywana funkcjami
       sklejanymi 3-go stopnia, w zewnetrznych - funkcjami liniowymi.
   W danych temperatura jest podawana w stopniach Celsjusza,
   natomiast wewnetrznie w programie uzywane sa kelwiny.
*/

/* Parametry funkcji przesuniecia profilu temperatury (funkcja Gaussa):
   d0: odleglosc centrum rozkladu
   a0: szerokosc polowkowa rozkladu,
   hs: maksymalne przesuniecie (wspolczynnik).
   Temperature na wysokosci h liczymy z wzoru T(h)=T(h+dh), gdzie
   h = hs * exp(-((d-d0)/a0)**2)
 */

static double deltah(double *rsd) {
	double x[3], d, dh;
	if (hs == 0.)
		return 0.;
	g2k(rsd, x);
	aminusb3(xref,x,x);
	d = sqrt(adotb3(x,x));
	d = (d - d0) / a0;
	dh = hs * exp(-pow(d*d,(double) ndh));
	return dh;
}

void deltah_init(double *x, double d, double a, double h, int n) {
	d0 = d * 1000.;
	a0 = a * 1000.;
	hs = h;
	dcopy3(x, xref);
	ndh = n;
}

/* Opis zaleznosci cisnienia i temperatury od wysokosci. */
void temph(double *h, double *t) {
	double hd;
	hd = h[0] + deltah(h+1);
	switch (tmodel) {
	case 1:
		t[0] = t0 + tgrad * hd;
		t[1] = tgrad;
		break;
	case 2:
		fpint(hd, t);
		break;
	default:
		fprintf(stderr,"model zaleznosci T od h ma numer %d,\n",tmodel);
		blad("a o takim modelu nic mi nie wiadomo.");
	}
	if (t[0] <= 0.) {
		fprintf(stderr,"Na wysokosci %lg m T=%lg K\n",h[0],t[0]);
		blad("temperatura w kelwinach musi byc dodatnia.");
	}
}

/* Wczytanie liczby punktow, wysokosci w nich  i wartosci temperatury. */
static void thinit() {
	double th1, th2;	/* Pochodne temperatury po wysokosci */
	double *h, *t;
	int i, np;
	if (verb) {
		printf("Podaj gradient temperatury ponizej modelowanego obszaru,\n");
		printf("liczbe punktow pomiarowych w obszarze modelowanym\n");
		printf("i gradient temperatury powyzej tego obszaru.\n");
	}
	scanf("%lg %d %lg", &th1, &np, &th2);
	h = (double *) xmalloc(sizeof(double)*(2*np));
	t = h + np;
	if (verb)
		printf("Podaj wartosci wysokosci i temperatur w %d punktach:\n",np);
	for (i = 0; i < np; i++) {
		scanf("%lg %lg",&h[i],&t[i]);
		/* Przeliczamy na skale Kelvina */
		t[i] += 273.15;
		if (t[i] <= 0.)
			blad("Temperatura musi byc dodatnia.");
	}
	splinit(h, t, np, th1, th2);
}

/* Wczytanie parametrow modeli opisu zaleznosci
   temperatury i cisnienia od wysokosci. */

void tinit() {
	scanf("%d",&tmodel);
	switch (tmodel) {
	case 1:
		if (verb)
			printf("Podaj temperature na poziomie morza i jej pionowy gradient\n");
		scanf("%lg %lg",&t0,&tgrad);
		t0 += 273.15;
		if (t0 <= 0.)
			blad("Temperatura w kelwinach musi byc dodatnia.");
		break;
	case 2:
		thinit();
		break;
	default:
		blad("Nieznany model zaleznosci temperatury od wysokosci.");
	}
}
